"""
Tests for linprog_simplex

"""
import numpy as np
from numpy.testing import assert_, assert_allclose, assert_equal

from quantecon.optimize import linprog_simplex


# Assert functions from scipy/optimize/tests/test_linprog.py

def _assert_infeasible(res):
    # res: linprog result object
    assert_(not res.success, "incorrectly reported success")
    assert_equal(res.status, 2, "failed to report infeasible status")


def _assert_unbounded(res):
    # res: linprog result object
    assert_(not res.success, "incorrectly reported success")
    assert_equal(res.status, 3, "failed to report unbounded status")


def _assert_success(res, c, b_ub=np.array([]), b_eq=np.array([]),
                    desired_fun=None, desired_x=None, rtol=1e-15, atol=1e-15):
    if not res.success:
        msg = "linprog status {0}".format(res.status)
        raise AssertionError(msg)

    assert_equal(res.status, 0)
    if desired_fun is not None:
        assert_allclose(res.fun, desired_fun,
                        err_msg="converged to an unexpected objective value",
                        rtol=rtol, atol=atol)
    if desired_x is not None:
        assert_allclose(res.x, desired_x,
                        err_msg="converged to an unexpected solution",
                        rtol=rtol, atol=atol)

    fun_p = c @ res.x
    fun_d = np.concatenate([b_ub, b_eq]) @ res.lambd
    assert_allclose(fun_p, fun_d, rtol=rtol, atol=atol)
    assert_allclose(res.fun, fun_d, rtol=rtol, atol=atol)


class TestLinprogSimplexScipy:
    # Test cases from scipy/optimize/tests/test_linprog.py

    def test_infeasible(self):
        # Test linprog response to an infeasible problem
        c = np.array([-1, -1]) * (-1)
        A_ub = [[1, 0],
                [0, 1],
                [-1, -1]]
        b_ub = [2, 2, -5]
        c, A_ub, b_ub = map(np.asarray, [c, A_ub, b_ub])
        res = linprog_simplex(c, A_ub=A_ub, b_ub=b_ub)
        _assert_infeasible(res)

    def test_unbounded(self):
        # Test linprog response to an unbounded problem
        c = np.array([1, 1])
        A_ub = [[-1, 1],
                [-1, -1]]
        b_ub = [-1, -2]
        c, A_ub, b_ub = map(np.asarray, [c, A_ub, b_ub])
        res = linprog_simplex(c, A_ub=A_ub, b_ub=b_ub)
        _assert_unbounded(res)

    def test_nontrivial_problem(self):
        # Test linprog for a problem involving all constraint types,
        # negative resource limits, and rounding issues.
        c = np.array([-1, 8, 4, -6]) * (-1)
        A_ub = [[-7, -7, 6, 9],
                [1, -1, -3, 0],
                [10, -10, -7, 7],
                [6, -1, 3, 4]]
        b_ub = [-3, 6, -6, 6]
        A_eq = [[-10, 1, 1, -8]]
        b_eq = [-4]
        c, A_ub, b_ub, A_eq, b_eq = \
            map(np.asarray, [c, A_ub, b_ub, A_eq, b_eq])
        desired_fun = 7083 / 1391 * (-1)
        desired_x = [101 / 1391, 1462 / 1391, 0, 752 / 1391]
        res = linprog_simplex(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq)
        _assert_success(res, c, b_ub=b_ub, b_eq=b_eq, desired_fun=desired_fun,
                        desired_x=desired_x)

    def test_network_flow(self):
        # A network flow problem with supply and demand at nodes
        # and with costs along directed edges.
        # https://www.princeton.edu/~rvdb/542/lectures/lec10.pdf
        c = np.array([2, 4, 9, 11, 4, 3, 8, 7, 0, 15, 16, 18]) * (-1)
        n, p = -1, 1
        A_eq = [
            [n, n, p, 0, p, 0, 0, 0, 0, p, 0, 0],
            [p, 0, 0, p, 0, p, 0, 0, 0, 0, 0, 0],
            [0, 0, n, n, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, p, p, 0, 0, p, 0],
            [0, 0, 0, 0, n, n, n, 0, p, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, n, n, 0, 0, p],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, n, n, n]]
        b_eq = [0, 19, -16, 33, 0, 0, -36]
        c, A_eq, b_eq = map(np.asarray, [c, A_eq, b_eq])
        desired_fun = -755
        res = linprog_simplex(c, A_eq=A_eq, b_eq=b_eq)
        _assert_success(res, c, b_eq=b_eq, desired_fun=desired_fun)

    def test_basic_artificial_vars(self):
        # Test if linprog succeeds when at the end of Phase 1 some artificial
        # variables remain basic, and the row in T corresponding to the
        # artificial variables is not all zero.
        c = np.array([-0.1, -0.07, 0.004, 0.004, 0.004, 0.004]) * (-1)
        A_ub = np.array([[1.0, 0, 0, 0, 0, 0], [-1.0, 0, 0, 0, 0, 0],
                         [0, -1.0, 0, 0, 0, 0], [0, 1.0, 0, 0, 0, 0],
                         [1.0, 1.0, 0, 0, 0, 0]])
        b_ub = np.array([3.0, 3.0, 3.0, 3.0, 20.0])
        A_eq = np.array([[1.0, 0, -1, 1, -1, 1], [0, -1.0, -1, 1, -1, 1]])
        b_eq = np.array([0, 0])
        res = linprog_simplex(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq)
        _assert_success(res, c, b_ub=b_ub, b_eq=b_eq, desired_fun=0,
                        desired_x=np.zeros_like(c))

    def test_bug_5400(self):
        # https://github.com/scipy/scipy/issues/5400
        f = 1 / 9
        g = -1e4
        h = -3.1
        A_ub = np.array([
            [1, -2.99, 0, 0, -3, 0, 0, 0, -1, -1, 0, -1, -1, 1, 1, 0, 0, 0, 0],
            [1, 0, -2.9, h, 0, -3, 0, -1, 0, 0, -1, 0, -1, 0, 0, 1, 1, 0, 0],
            [1, 0, 0, h, 0, 0, -3, -1, -1, 0, -1, -1, 0, 0, 0, 0, 0, 1, 1],
            [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1],
            [0, 1.99, -1, -1, 0, 0, 0, -1, f, f, 0, 0, 0, g, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 2, -1, -1, 0, 0, 0, -1, f, f, 0, g, 0, 0, 0, 0],
            [0, -1, 1.9, 2.1, 0, 0, 0, f, -1, -1, 0, 0, 0, 0, 0, g, 0, 0, 0],
            [0, 0, 0, 0, -1, 2, -1, 0, 0, 0, f, -1, f, 0, 0, 0, g, 0, 0],
            [0, -1, -1, 2.1, 0, 0, 0, f, f, -1, 0, 0, 0, 0, 0, 0, 0, g, 0],
            [0, 0, 0, 0, -1, -1, 2, 0, 0, 0, f, f, -1, 0, 0, 0, 0, 0, g]])

        b_ub = np.array([
            0.0, 0, 0, 100, 100, 100, 100, 100, 100, 900, 900, 900, 900, 900,
            900, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

        c = np.array([-1.0, 1, 1, 1, 1, 1, 1, 1, 1,
                      1, 1, 1, 1, 0, 0, 0, 0, 0, 0]) * (-1)

        desired_fun = 106.63507541835018

        res = linprog_simplex(c, A_ub=A_ub, b_ub=b_ub)
        _assert_success(res, c, b_ub=b_ub, desired_fun=desired_fun,
                        rtol=1e-8, atol=1e-8)

    def test_issue_8174_stackoverflow(self):
        # Test supplementary example from issue 8174.
        # https://github.com/scipy/scipy/issues/8174
        # https://stackoverflow.com/questions/47717012/
        c = np.array([1, 0, 0, 0, 0, 0, 0]) * (-1)
        A_ub = -np.identity(7)
        b_ub = np.full(7, -2)
        A_eq = np.array([
            [1, 1, 1, 1, 1, 1, 0],
            [0.3, 1.3, 0.9, 0, 0, 0, -1],
            [0.3, 0, 0, 0, 0, 0, -2/3],
            [0, 0.65, 0, 0, 0, 0, -1/15],
            [0, 0, 0.3, 0, 0, 0, -1/15]
        ])
        b_eq = np.array([100, 0, 0, 0, 0])

        res = linprog_simplex(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq)
        _assert_success(res, c, b_ub=b_ub, b_eq=b_eq)

    def test_linprog_cyclic_recovery(self):
        # Test linprogs recovery from cycling using the Klee-Minty problem
        # Klee-Minty  https://www.math.ubc.ca/~israel/m340/kleemin3.pdf
        c = np.array([100, 10, 1])
        A_ub = [[1, 0, 0],
                [20, 1, 0],
                [200, 20, 1]]
        b_ub = [1, 100, 10000]
        c, A_ub, b_ub = map(np.asarray, [c, A_ub, b_ub])
        desired_x = [0, 0, 10000]
        res = linprog_simplex(c, A_ub=A_ub, b_ub=b_ub)
        _assert_success(res, c, b_ub=b_ub, desired_x=desired_x)

    def test_linprog_cyclic_bland(self):
        # Test the effect of Bland's rule on a cycling problem
        c = np.array([-10, 57, 9, 24.]) * (-1)
        A_ub = np.array([[0.5, -5.5, -2.5, 9],
                         [0.5, -1.5, -0.5, 1],
                         [1, 0, 0, 0]])
        b_ub = np.array([0, 0, 1])
        desired_x = [1, 0, 1, 0]
        res = linprog_simplex(c, A_ub=A_ub, b_ub=b_ub)
        _assert_success(res, c, b_ub=b_ub, desired_x=desired_x)

    def test_linprog_cyclic_bland_bug_8561(self):
        # Test that pivot row is chosen correctly when using Bland's rule
        c = np.array([7, 0, -4, 1.5, 1.5]) * (-1)
        A_ub = np.array([
            [4, 5.5, 1.5, 1.0, -3.5],
            [1, -2.5, -2, 2.5, 0.5],
            [3, -0.5, 4, -12.5, -7],
            [-1, 4.5, 2, -3.5, -2],
            [5.5, 2, -4.5, -1, 9.5]])
        b_ub = np.array([0, 0, 0, 0, 1])
        # desired_x = [0, 0, 19, 16/3, 29/3]
        res = linprog_simplex(c, A_ub=A_ub, b_ub=b_ub)
        _assert_success(res, c, b_ub=b_ub)

    def test_bug_8662(self):
        # linprog simplex used to report inncorrect optimal results
        # https://github.com/scipy/scipy/issues/8662
        c = np.array([-10, 10, 6, 3]) * (-1)
        A_ub = [[8, -8, -4, 6],
                [-8, 8, 4, -6],
                [-4, 4, 8, -4],
                [3, -3, -3, -10]]
        b_ub = [9, -9, -9, -4]
        c, A_ub, b_ub = map(np.asarray, [c, A_ub, b_ub])
        desired_fun = -36.0000000000
        res = linprog_simplex(c, A_ub=A_ub, b_ub=b_ub)
        _assert_success(res, c, b_ub=b_ub, desired_fun=desired_fun)



# flake8: noqa
"""
Import the main names to top level.
"""

__version__ = '0.7.0'

try:
    import numba
except:
    raise ImportError(
        "Cannot import numba from current anaconda distribution. \
            Please run `conda install numba` to install the latest version.")

#-Modules-#
from . import distributions
from . import game_theory
from . import quad
from . import random
from . import optimize

#-Objects-#
from ._compute_fp import compute_fixed_point
from ._discrete_rv import DiscreteRV
from ._dle import DLE
from ._ecdf import ECDF
from ._estspec import smooth, periodogram, ar_periodogram
# from .game_theory import <objects-here>                           #Place Holder if we wish to promote any general objects to the qe namespace.
from ._graph_tools import DiGraph, random_tournament_graph
from ._gridtools import (
    cartesian, mlinspace, cartesian_nearest_index, simplex_grid, simplex_index,
    num_compositions
)
from ._inequality import lorenz_curve, gini_coefficient, shorrocks_index, \
    rank_size
from ._kalman import Kalman
from ._lae import LAE
from ._arma import ARMA
from ._lqcontrol import LQ, LQMarkov
from ._filter import hamilton_filter
from ._lqnash import nnash
from ._ivp import IVP
from ._lss import LinearStateSpace
from ._matrix_eqn import solve_discrete_lyapunov, solve_discrete_riccati
from ._quadsums import var_quadratic_sum, m_quadratic_sum
#->Propose Delete From Top Level
#Promote to keep current examples working
from .markov import MarkovChain, random_markov_chain, random_stochastic_matrix, \
    gth_solve, tauchen, rouwenhorst, estimate_mc
#Imports that Should be Deprecated with markov package
from .markov import mc_compute_stationary, mc_sample_path
#<-
from ._rank_nullspace import rank_est, nullspace
from ._robustlq import RBLQ
from .util import searchsorted, fetch_nb_dependencies, tic, tac, toc



"""
Generates an array of draws from a discrete random variable with a
specified vector of probabilities.

"""

import numpy as np
from .util import check_random_state


class DiscreteRV:
    """
    Generates an array of draws from a discrete random variable with
    vector of probabilities given by q.

    Parameters
    ----------
    q : array_like(float)
        Nonnegative numbers that sum to 1.

    Attributes
    ----------
    q : see Parameters.
    Q : array_like(float)
        The cumulative sum of q.

    """

    def __init__(self, q):
        self._q = np.asarray(q)
        self.Q = np.cumsum(q)

    def __repr__(self):
        return "DiscreteRV with {n} elements".format(n=self._q.size)

    def __str__(self):
        return self.__repr__()

    @property
    def q(self):
        """
        Getter method for q.

        """
        return self._q

    @q.setter
    def q(self, val):
        """
        Setter method for q.

        """
        self._q = np.asarray(val)
        self.Q = np.cumsum(val)

    def draw(self, k=1, random_state=None):
        """
        Returns k draws from q.

        For each such draw, the value i is returned with probability
        q[i].

        Parameters
        ----------
        k : scalar(int), optional
            Number of draws to be returned

        random_state : int or np.random.RandomState/Generator, optional
            Random seed (integer) or np.random.RandomState or Generator
            instance to set the initial state of the random number
            generator for reproducibility. If None, a randomly
            initialized RandomState is used.

        Returns
        -------
        array_like(int)
            An array of k independent draws from q

        """
        random_state = check_random_state(random_state)

        return self.Q.searchsorted(random_state.uniform(0, 1, size=k),
                                   side='right')



"""
Tests for Array Utilities

Functions
---------
searchsorted

"""
import numpy as np
from numpy.testing import assert_
from quantecon.util import searchsorted


def test_searchsorted():
    a = np.array([0.2, 0.4, 1.0])
    assert_(searchsorted(a, 0.1) == 0)
    assert_(searchsorted(a, 0.4) == 2)
    assert_(searchsorted(a, 2) == 3)

    a = np.ones(0)
    for (v, i) in zip([0, 1, 2], [0, 0, 0]):
        assert_(searchsorted(a, v) == i)

    a = np.ones(1)
    for (v, i) in zip([0, 1, 2], [0, 1, 1]):
        assert_(searchsorted(a, v) == i)

    a = np.ones(2)
    for (v, i) in zip([0, 1, 2], [0, 2, 2]):
        assert_(searchsorted(a, v) == i)



"""
Tests for quantecon.util

"""
import numpy as np
from numpy.testing import assert_allclose
from quantecon import _matrix_eqn as qme


def test_solve_discrete_lyapunov_zero():
    'Simple test where X is all zeros'
    A = np.eye(4) * .95
    B = np.zeros((4, 4))

    X = qme.solve_discrete_lyapunov(A, B)

    assert_allclose(X, np.zeros((4, 4)))


def test_solve_discrete_lyapunov_B():
    'Simple test where X is same as B'
    A = np.ones((2, 2)) * .5
    B = np.array([[.5, -.5], [-.5, .5]])

    X = qme.solve_discrete_lyapunov(A, B)

    assert_allclose(B, X)

def test_solve_discrete_lyapunov_complex():
    'Complex test, A is companion matrix'
    A = np.array([[0.5 + 0.3j, 0.1 + 0.1j],
                  [         1,          0]])
    B = np.eye(2)

    X = qme.solve_discrete_lyapunov(A, B)

    assert_allclose(np.dot(np.dot(A, X), A.conj().transpose()) - X, -B,
                    atol=1e-15)



"""
Tests for ecdf.py

"""
import numpy as np
from numpy.testing import assert_, assert_allclose
from quantecon import ECDF


class TestECDF:

    @classmethod
    def setup_method(cls):
        cls.obs = np.random.rand(40)  # observations defining dist
        cls.ecdf = ECDF(cls.obs)

    def test_call_high(self):
        "ecdf: x above all obs give 1.0"
        # all of self.obs <= 1 so ecdf(1.1) should be 1
        assert_allclose(self.ecdf(1.1), 1.0)

    def test_call_low(self):
        "ecdf: x below all obs give 0.0"
        # all of self.obs <= 1 so ecdf(1.1) should be 1
        assert_allclose(self.ecdf(-0.1), 0.0)

    def test_ascending(self):
        "ecdf: larger values should return F(x) at least as big"
        x = np.random.rand()
        F_1 = self.ecdf(x)
        F_2 = self.ecdf(1.1 * x)
        assert_(F_2 >= F_1)

    def test_vectorized(self):
        "ecdf: testing vectorized __call__ method"
        t = np.linspace(-1, 1, 100)
        e = self.ecdf(t)
        assert_(t.shape == e.shape)
        assert_(e.dtype == float)
        t = np.linspace(-1, 1, 100).reshape(2, 2, 25)
        e = self.ecdf(t)
        assert_(t.shape == e.shape)
        assert_(e.dtype == float)



"""
Utilities to support Numba jitted functions

"""
import numpy as np
from numba import jit, types
from numba.extending import overload
from numba.np.linalg import _LAPACK


# BLAS kinds as letters
_blas_kinds = {
    types.float32: 's',
    types.float64: 'd',
    types.complex64: 'c',
    types.complex128: 'z',
}


def _numba_linalg_solve(a, b):  # pragma: no cover
    pass


@overload(_numba_linalg_solve, jit_options={'cache':True})
def _numba_linalg_solve_ol(a, b):
    """
    Solve the linear equation ax = b directly calling a Numba internal
    function. The data in `a` and `b` are interpreted in Fortran order,
    and dtype of `a` and `b` must be the same, one of {float32, float64,
    complex64, complex128}. `a` and `b` are modified in place, and the
    solution is stored in `b`. *No error check is made for the inputs.*
    Only work in a Numba-jitted function.

    Parameters
    ----------
    a : ndarray(ndim=2)
        2-dimensional ndarray of shape (n, n).

    b : ndarray(ndim=1 or 2)
        1-dimensional ndarray of shape (n,) or 2-dimensional ndarray of
        shape (n, nrhs).

    Returns
    -------
    r : scalar(int)
        r = 0 if successful.

    Notes
    -----
    From github.com/numba/numba/blob/main/numba/np/linalg.py

    """
    numba_xgesv = _LAPACK().numba_xgesv(a.dtype)
    kind = ord(_blas_kinds[a.dtype])

    def _numba_linalg_solve_impl(a, b):  # pragma: no cover
        n = a.shape[-1]
        if b.ndim == 1:
            nrhs = 1
        else:  # b.ndim == 2
            nrhs = b.shape[-1]
        F_INT_nptype = np.int32
        ipiv = np.empty(n, dtype=F_INT_nptype)

        r = numba_xgesv(
            kind,         # kind
            n,            # n
            nrhs,         # nhrs
            a.ctypes,     # a
            n,            # lda
            ipiv.ctypes,  # ipiv
            b.ctypes,     # b
            n             # ldb
        )
        return r

    return _numba_linalg_solve_impl


@jit(types.intp(types.intp, types.intp), nopython=True, cache=True)
def comb_jit(N, k):
    """
    Numba jitted function that computes N choose k. Return `0` if the
    outcome exceeds the maximum value of `np.intp` or if N < 0, k < 0,
    or k > N.

    Parameters
    ----------
    N : scalar(int)

    k : scalar(int)

    Returns
    -------
    val : scalar(int)

    """
    # From scipy.special._comb_int_long
    # github.com/scipy/scipy/blob/v1.0.0/scipy/special/_comb.pyx
    INTP_MAX = np.iinfo(np.intp).max
    if N < 0 or k < 0 or k > N:
        return 0
    if k == 0:
        return 1
    if k == 1:
        return N
    if N == INTP_MAX:
        return 0

    M = N + 1
    nterms = min(k, N - k)

    val = 1

    for j in range(1, nterms + 1):
        # Overflow check
        if val > INTP_MAX // (M - j):
            return 0

        val *= M - j
        val //= j

    return val



"""
Tests for ricatti.py

"""
import numpy as np
from numpy.testing import assert_allclose
from quantecon import solve_discrete_lyapunov


def test_dlyap_simple_ones():
    A = np.zeros((4, 4))
    B = np.ones((4, 4))

    sol = solve_discrete_lyapunov(A, B)

    assert_allclose(sol, np.ones((4, 4)))


def test_dlyap_scalar():
    a = .5
    b = .75

    sol = solve_discrete_lyapunov(a, b)

    assert_allclose(sol, np.ones((1, 1)))



"""
Tests for repeated_game.py

"""
import numpy as np
from numpy.testing import assert_allclose
from quantecon.game_theory import NormalFormGame, RepeatedGame


class TestAS():
    def setup_method(self):
        self.game_dicts = []

        # Example from Abreu and Sannikov (2014)
        bimatrix = [[(16, 9), (3, 13), (0, 3)],
                    [(21, 1), (10, 4), (-1, 0)],
                    [(9, 0), (5, -4), (-5, -15)]]
        vertices = np.array([[7.33770472e+00, 1.09826253e+01],
                             [1.12568240e+00, 2.80000000e+00],
                             [7.33770472e+00, 4.44089210e-16],
                             [7.86308964e+00, 4.44089210e-16],
                             [1.97917977e+01, 2.80000000e+00],
                             [1.55630896e+01, 9.10000000e+00]])
        d = {'sg': NormalFormGame(bimatrix),
             'delta': 0.3,
             'vertices': vertices,
             'u': np.zeros(2)}
        self.game_dicts.append(d)

        # Prisoner's dilemma
        bimatrix = [[(9, 9), (1, 10)],
                    [(10, 1), (3, 3)]]
        vertices = np.array([[3.  , 3.  ],
                             [9.75, 3.  ],
                             [9.  , 9.  ],
                             [3.  , 9.75]])
        d = {'sg': NormalFormGame(bimatrix),
             'delta': 0.9,
             'vertices': vertices,
             'u': np.array([3., 3.])}
        self.game_dicts.append(d)

    def test_abreu_sannikov(self):
        for d in self.game_dicts:
            rpg = RepeatedGame(d['sg'], d['delta'])
            for method in ('abreu_sannikov', 'AS'):
                hull = rpg.equilibrium_payoffs(method=method,
                                               options={'u_init': d['u']})
                assert_allclose(hull.points[hull.vertices], d['vertices'])



r"""
This file contains some useful objects for handling a finite-state
discrete-time Markov chain.

Definitions and Some Basic Facts about Markov Chains
----------------------------------------------------

Let :math:`\{X_t\}` be a Markov chain represented by an :math:`n \times
n` stochastic matrix :math:`P`. State :math:`i` *has access* to state
:math:`j`, denoted :math:`i \to j`, if :math:`i = j` or :math:`P^k[i, j]
> 0` for some :math:`k = 1, 2, \ldots`; :math:`i` and `j` *communicate*,
denoted :math:`i \leftrightarrow j`, if :math:`i \to j` and :math:`j \to
i`. The binary relation :math:`\leftrightarrow` is an equivalent
relation. A *communication class* of the Markov chain :math:`\{X_t\}`,
or of the stochastic matrix :math:`P`, is an equivalent class of
:math:`\leftrightarrow`. Equivalently, a communication class is a
*strongly connected component* (SCC) in the associated *directed graph*
:math:`\Gamma(P)`, a directed graph with :math:`n` nodes where there is
an edge from :math:`i` to :math:`j` if and only if :math:`P[i, j] > 0`.
The Markov chain, or the stochastic matrix, is *irreducible* if it
admits only one communication class, or equivalently, if
:math:`\Gamma(P)` is *strongly connected*.

A state :math:`i` is *recurrent* if :math:`i \to j` implies :math:`j \to
i`; it is *transient* if it is not recurrent. For any :math:`i, j`
contained in a communication class, :math:`i` is recurrent if and only
if :math:`j` is recurrent. Therefore, recurrence is a property of a
communication class. Thus, a communication class is a *recurrent class*
if it contains a recurrent state. Equivalently, a recurrent class is a
SCC that corresponds to a sink node in the *condensation* of the
directed graph :math:`\Gamma(P)`, where the condensation of
:math:`\Gamma(P)` is a directed graph in which each SCC is replaced with
a single node and there is an edge from one SCC :math:`C` to another SCC
:math:`C'` if :math:`C \neq C'` and there is an edge from some node in
:math:`C` to some node in :math:`C'`. A recurrent class is also called a
*closed communication class*. The condensation is acyclic, so that there
exists at least one recurrent class.

For example, if the entries of :math:`P` are all strictly positive, then
the whole state space is a communication class as well as a recurrent
class. (More generally, if there is only one communication class, then
it is a recurrent class.) As another example, consider the stochastic
matrix :math:`P = [[1, 0], [0,5, 0.5]]`. This has two communication
classes, :math:`\{0\}` and :math:`\{1\}`, and :math:`\{0\}` is the only
recurrent class.

A *stationary distribution* of the Markov chain :math:`\{X_t\}`, or of
the stochastic matrix :math:`P`, is a nonnegative vector :math:`x` such
that :math:`x' P = x'` and :math:`x' \mathbf{1} = 1`, where
:math:`\mathbf{1}` is the vector of ones. The Markov chain has a unique
stationary distribution if and only if it has a unique recurrent class.
More generally, each recurrent class has a unique stationary
distribution whose support equals that recurrent class. The set of all
stationary distributions is given by the convex hull of these unique
stationary distributions for the recurrent classes.

A natural number :math:`d` is the *period* of state :math:`i` if it is
the greatest common divisor of all :math:`k`'s such that :math:`P^k[i,
i] > 0`; equivalently, it is the GCD of the lengths of the cycles in
:math:`\Gamma(P)` passing through :math:`i`. For any :math:`i, j`
contained in a communication class, :math:`i` has period :math:`d` if
and only if :math:`j` has period :math:`d`. The *period* of an
irreducible Markov chain (or of an irreducible stochastic matrix) is the
period of any state. We define the period of a general (not necessarily
irreducible) Markov chain to be the least common multiple of the periods
of its recurrent classes, where the period of a recurrent class is the
period of any state in that class. A Markov chain is *aperiodic* if its
period is one. A Markov chain is irreducible and aperiodic if and only
if it is *uniformly ergodic*, i.e., there exists some :math:`m` such
that :math:`P^m[i, j] > 0` for all :math:`i, j` (in this case, :math:`P`
is also called *primitive*).

Suppose that an irreducible Markov chain has period :math:`d`. Fix any
state, say state :math:`0`. For each :math:`m = 0, \ldots, d-1`, let
:math:`S_m` be the set of states :math:`i` such that :math:`P^{kd+m}[0,
i] > 0` for some :math:`k`. These sets :math:`S_0, \ldots, S_{d-1}`
constitute a partition of the state space and are called the *cyclic
classes*. For each :math:`S_m` and each :math:`i \in S_m`, we have
:math:`\sum_{j \in S_{m+1}} P[i, j] = 1`, where :math:`S_d = S_0`.

"""
import numbers
from math import gcd
import numpy as np
from scipy import sparse
from numba import jit

from .gth_solve import gth_solve
from .._graph_tools import DiGraph
from ..util import searchsorted, check_random_state, rng_integers


class MarkovChain:
    """
    Class for a finite-state discrete-time Markov chain. It stores
    useful information such as the stationary distributions, and
    communication, recurrent, and cyclic classes, and allows simulation
    of state transitions.

    Parameters
    ----------
    P : array_like or scipy sparse matrix (float, ndim=2)
        The transition matrix.  Must be of shape n x n.

    state_values : array_like(default=None)
        Array_like of length n containing the values associated with the
        states, which must be homogeneous in type. If None, the values
        default to integers 0 through n-1.

    Attributes
    ----------
    P : ndarray or scipy.sparse.csr_matrix (float, ndim=2)
        See Parameters

    stationary_distributions : array_like(float, ndim=2)
        Array containing stationary distributions, one for each
        recurrent class, as rows.

    is_irreducible : bool
        Indicate whether the Markov chain is irreducible.

    num_communication_classes : int
        The number of the communication classes.

    communication_classes_indices : list(ndarray(int))
        List of numpy arrays containing the indices of the communication
        classes.

    communication_classes : list(ndarray)
        List of numpy arrays containing the communication classes, where
        the states are annotated with their values (if `state_values` is
        not None).

    num_recurrent_classes : int
        The number of the recurrent classes.

    recurrent_classes_indices : list(ndarray(int))
        List of numpy arrays containing the indices of the recurrent
        classes.

    recurrent_classes : list(ndarray)
        List of numpy arrays containing the recurrent classes, where the
        states are annotated with their values (if `state_values` is not
        None).

    is_aperiodic : bool
        Indicate whether the Markov chain is aperiodic.

    period : int
        The period of the Markov chain.

    cyclic_classes_indices : list(ndarray(int))
        List of numpy arrays containing the indices of the cyclic
        classes. Defined only when the Markov chain is irreducible.

    cyclic_classes : list(ndarray)
        List of numpy arrays containing the cyclic classes, where the
        states are annotated with their values (if `state_values` is not
        None). Defined only when the Markov chain is irreducible.

    Notes
    -----
    In computing stationary distributions, if the input matrix is a
    sparse matrix, internally it is converted to a dense matrix.

    """

    def __init__(self, P, state_values=None):
        if sparse.issparse(P):  # Sparse matrix
            self.P = sparse.csr_matrix(P)
            self.is_sparse = True
        else:  # Dense matrix
            self.P = np.asarray(P)
            self.is_sparse = False

        # Check Properties
        # Double check that P is a square matrix
        if len(self.P.shape) != 2 or self.P.shape[0] != self.P.shape[1]:
            raise ValueError('P must be a square matrix')

        # The number of states
        self.n = self.P.shape[0]

        # Double check that P is a nonnegative matrix
        if not self.is_sparse:
            data_nonnegative = (self.P >= 0)  # ndarray
        else:
            data_nonnegative = (self.P.data >= 0)  # csr_matrx
        if not np.all(data_nonnegative):
            raise ValueError('P must be nonnegative')

        # Double check that the rows of P sum to one
        row_sums = self.P.sum(axis=1)
        if self.is_sparse:  # row_sums is np.matrix (ndim=2)
            row_sums = row_sums.getA1()
        if not np.allclose(row_sums, np.ones(self.n)):
            raise ValueError('The rows of P must sum to 1')

        # Call the setter method
        self.state_values = state_values

        # To analyze the structure of P as a directed graph
        self._digraph = None

        self._stationary_dists = None
        self._cdfs = None  # For dense matrix
        self._cdfs1d = None  # For sparse matrix

    def __repr__(self):
        msg = "Markov chain with transition matrix \nP = \n{0}"

        if self._stationary_dists is None:
            return msg.format(self.P)
        else:
            msg = msg + "\nand stationary distributions \n{1}"
            return msg.format(self.P, self._stationary_dists)

    def __str__(self):
        return str(self.__repr__)

    @property
    def state_values(self):
        return self._state_values

    @state_values.setter
    def state_values(self, values):
        if values is None:
            self._state_values = None
        else:
            values = np.asarray(values)
            if (values.ndim < 1) or (values.shape[0] != self.n):
                raise ValueError(
                    'state_values must be an array_like of length n'
                )
            if np.issubdtype(values.dtype, np.object_):
                raise ValueError(
                    'data in state_values must be homogeneous in type'
                )
            self._state_values = values

    def get_index(self, value):
        """
        Return the index (or indices) of the given value (or values) in
        `state_values`.

        Parameters
        ----------
        value
            Value(s) to get the index (indices) for.

        Returns
        -------
        idx : int or ndarray(int)
            Index of `value` if `value` is a single state value; array
            of indices if `value` is an array_like of state values.

        """
        if self.state_values is None:
            state_values_ndim = 1
        else:
            state_values_ndim = self.state_values.ndim

        values = np.asarray(value)

        if values.ndim <= state_values_ndim - 1:
            return self._get_index(value)
        elif values.ndim == state_values_ndim:  # array of values
            k = values.shape[0]
            idx = np.empty(k, dtype=int)
            for i in range(k):
                idx[i] = self._get_index(values[i])
            return idx
        else:
            raise ValueError('invalid value')

    def _get_index(self, value):
        """
        Return the index of the given value in `state_values`.

        Parameters
        ----------
        value
            Value to get the index for.

        Returns
        -------
        idx : int
            Index of `value`.

        """
        error_msg = 'value {0} not found'.format(value)

        if self.state_values is None:
            if isinstance(value, numbers.Integral) and (0 <= value < self.n):
                return value
            else:
                raise ValueError(error_msg)

        # if self.state_values is not None:
        if self.state_values.ndim == 1:
            try:
                idx = np.where(self.state_values == value)[0][0]
                return idx
            except IndexError:
                raise ValueError(error_msg)
        else:
            idx = 0
            while idx < self.n:
                if np.array_equal(self.state_values[idx], value):
                    return idx
                idx += 1
            raise ValueError(error_msg)

    @property
    def digraph(self):
        if self._digraph is None:
            self._digraph = DiGraph(self.P, node_labels=self.state_values)
        return self._digraph

    @property
    def is_irreducible(self):
        return self.digraph.is_strongly_connected

    @property
    def num_communication_classes(self):
        return self.digraph.num_strongly_connected_components

    @property
    def communication_classes_indices(self):
        return self.digraph.strongly_connected_components_indices

    @property
    def communication_classes(self):
        return self.digraph.strongly_connected_components

    @property
    def num_recurrent_classes(self):
        return self.digraph.num_sink_strongly_connected_components

    @property
    def recurrent_classes_indices(self):
        return self.digraph.sink_strongly_connected_components_indices

    @property
    def recurrent_classes(self):
        return self.digraph.sink_strongly_connected_components

    @property
    def is_aperiodic(self):
        if self.is_irreducible:
            return self.digraph.is_aperiodic
        else:
            return self.period == 1

    @property
    def period(self):
        if self.is_irreducible:
            return self.digraph.period
        else:
            # Determine the period, the LCM of the periods of rec_classes
            d = 1
            for rec_class in self.recurrent_classes_indices:
                period = self.digraph.subgraph(rec_class).period
                d = (d * period) // gcd(d, period)

            return d

    @property
    def cyclic_classes(self):
        if not self.is_irreducible:
            raise NotImplementedError(
                'Not defined for a reducible Markov chain'
            )
        else:
            return self.digraph.cyclic_components

    @property
    def cyclic_classes_indices(self):
        if not self.is_irreducible:
            raise NotImplementedError(
                'Not defined for a reducible Markov chain'
            )
        else:
            return self.digraph.cyclic_components_indices

    def _compute_stationary(self):
        """
        Store the stationary distributions in self._stationary_distributions.

        """
        if self.is_irreducible:
            if not self.is_sparse:  # Dense
                stationary_dists = gth_solve(self.P).reshape(1, self.n)
            else:  # Sparse
                stationary_dists = \
                    gth_solve(self.P.toarray(),
                              overwrite=True).reshape(1, self.n)
        else:
            rec_classes = self.recurrent_classes_indices
            stationary_dists = np.zeros((len(rec_classes), self.n))
            for i, rec_class in enumerate(rec_classes):
                P_rec_class = self.P[np.ix_(rec_class, rec_class)]
                if self.is_sparse:
                    P_rec_class = P_rec_class.toarray()
                stationary_dists[i, rec_class] = \
                    gth_solve(P_rec_class, overwrite=True)

        self._stationary_dists = stationary_dists

    @property
    def stationary_distributions(self):
        if self._stationary_dists is None:
            self._compute_stationary()
        return self._stationary_dists

    @property
    def cdfs(self):
        if (self._cdfs is None) and not self.is_sparse:
            # See issue #137#issuecomment-96128186
            cdfs = np.empty((self.n, self.n), order='C', dtype=self.P.dtype)
            np.cumsum(self.P, axis=-1, out=cdfs)
            self._cdfs = cdfs
        return self._cdfs

    @property
    def cdfs1d(self):
        if (self._cdfs1d is None) and self.is_sparse:
            data = self.P.data
            indptr = self.P.indptr

            cdfs1d = np.empty(self.P.nnz, order='C', dtype=data.dtype)
            for i in range(self.n):
                cdfs1d[indptr[i]:indptr[i+1]] = \
                    data[indptr[i]:indptr[i+1]].cumsum()
            self._cdfs1d = cdfs1d
        return self._cdfs1d

    def simulate_indices(self, ts_length, init=None, num_reps=None,
                         random_state=None):
        """
        Simulate time series of state transitions, where state indices
        are returned.

        Parameters
        ----------
        ts_length : scalar(int)
            Length of each simulation.

        init : int or array_like(int, ndim=1), optional
            Initial state(s). If None, the initial state is randomly
            drawn.

        num_reps : scalar(int), optional(default=None)
            Number of repetitions of simulation.

        random_state : int or np.random.RandomState/Generator, optional
            Random seed (integer) or np.random.RandomState or Generator
            instance to set the initial state of the random number
            generator for reproducibility. If None, a randomly
            initialized RandomState is used.

        Returns
        -------
        X : ndarray(ndim=1 or 2)
            Array containing the state values of the sample path(s). See
            the `simulate` method for more information.

        """
        random_state = check_random_state(random_state)
        dim = 1  # Dimension of the returned array: 1 or 2

        msg_out_of_range = 'index {init} is out of the state space'

        try:
            k = len(init)  # init is an array
            dim = 2
            init_states = np.asarray(init, dtype=int)
            # Check init_states are in the state space
            if (init_states >= self.n).any() or (init_states < -self.n).any():
                idx = np.where(
                    (init_states >= self.n) + (init_states < -self.n)
                )[0][0]
                raise ValueError(msg_out_of_range.format(init=idx))
            if num_reps is not None:
                k *= num_reps
                init_states = np.tile(init_states, num_reps)
        except TypeError:  # init is a scalar(int) or None
            k = 1
            if num_reps is not None:
                dim = 2
                k = num_reps
            if init is None:
                init_states = rng_integers(random_state, self.n, size=k)
            elif isinstance(init, numbers.Integral):
                # Check init is in the state space
                if init >= self.n or init < -self.n:
                    raise ValueError(msg_out_of_range.format(init=init))
                init_states = np.ones(k, dtype=int) * init
            else:
                raise ValueError(
                    'init must be int, array_like of ints, or None'
                )

        # === set up array to store output === #
        X = np.empty((k, ts_length), dtype=int)

        # Random values, uniformly sampled from [0, 1)
        random_values = random_state.random(size=(k, ts_length-1))

        # Generate sample paths and store in X
        if not self.is_sparse:  # Dense
            _generate_sample_paths(
                self.cdfs, init_states, random_values, out=X
            )
        else:  # Sparse
            _generate_sample_paths_sparse(
                self.cdfs1d, self.P.indices, self.P.indptr, init_states,
                random_values, out=X
            )

        if dim == 1:
            return X[0]
        else:
            return X

    def simulate(self, ts_length, init=None, num_reps=None, random_state=None):
        """
        Simulate time series of state transitions, where the states are
        annotated with their values (if `state_values` is not None).

        Parameters
        ----------
        ts_length : scalar(int)
            Length of each simulation.

        init : scalar or array_like, optional(default=None)
            Initial state values(s). If None, the initial state is
            randomly drawn.

        num_reps : scalar(int), optional(default=None)
            Number of repetitions of simulation.

        random_state : int or np.random.RandomState/Generator, optional
            Random seed (integer) or np.random.RandomState or Generator
            instance to set the initial state of the random number
            generator for reproducibility. If None, a randomly
            initialized RandomState is used.

        Returns
        -------
        X : ndarray(ndim=1 or 2)
            Array containing the sample path(s), of shape (ts_length,)
            if init is a scalar (integer) or None and num_reps is None;
            of shape (k, ts_length) otherwise, where k = len(init) if
            (init, num_reps) = (array, None), k = num_reps if (init,
            num_reps) = (int or None, int), and k = len(init)*num_reps
            if (init, num_reps) = (array, int).

        """
        if init is not None:
            init_idx = self.get_index(init)
        else:
            init_idx = None
        X = self.simulate_indices(ts_length, init=init_idx, num_reps=num_reps,
                                  random_state=random_state)

        # Annotate states
        if self.state_values is not None:
            X = self.state_values[X]

        return X


@jit(nopython=True)
def _generate_sample_paths(P_cdfs, init_states, random_values, out):
    """
    Generate num_reps sample paths of length ts_length, where num_reps =
    out.shape[0] and ts_length = out.shape[1].

    Parameters
    ----------
    P_cdfs : ndarray(float, ndim=2)
        Array containing as rows the CDFs of the state transition.

    init_states : array_like(int, ndim=1)
        Array containing the initial states. Its length must be equal to
        num_reps.

    random_values : ndarray(float, ndim=2)
        Array containing random values from [0, 1). Its shape must be
        equal to (num_reps, ts_length-1)

    out : ndarray(int, ndim=2)
        Array to store the sample paths.

    Notes
    -----
    This routine is jit-complied by Numba.

    """
    num_reps, ts_length = out.shape

    for i in range(num_reps):
        out[i, 0] = init_states[i]
        for t in range(ts_length-1):
            out[i, t+1] = searchsorted(P_cdfs[out[i, t]], random_values[i, t])


@jit(nopython=True)
def _generate_sample_paths_sparse(P_cdfs1d, indices, indptr, init_states,
                                  random_values, out):
    """
    For sparse matrix.

    Generate num_reps sample paths of length ts_length, where num_reps =
    out.shape[0] and ts_length = out.shape[1].

    Parameters
    ----------
    P_cdfs1d : ndarray(float, ndim=1)
        1D array containing the CDFs of the state transition.

    indices : ndarray(int, ndim=1)
        CSR format index array.

    indptr : ndarray(int, ndim=1)
        CSR format index pointer array.

    init_states : array_like(int, ndim=1)
        Array containing the initial states. Its length must be equal to
        num_reps.

    random_values : ndarray(float, ndim=2)
        Array containing random values from [0, 1). Its shape must be
        equal to (num_reps, ts_length-1)

    out : ndarray(int, ndim=2)
        Array to store the sample paths.

    Notes
    -----
    This routine is jit-complied by Numba.

    """
    num_reps, ts_length = out.shape

    for i in range(num_reps):
        out[i, 0] = init_states[i]
        for t in range(ts_length-1):
            k = searchsorted(P_cdfs1d[indptr[out[i, t]]:indptr[out[i, t]+1]],
                             random_values[i, t])
            out[i, t+1] = indices[indptr[out[i, t]]+k]


def mc_compute_stationary(P):
    """
    Computes stationary distributions of P, one for each recurrent
    class. Any stationary distribution is written as a convex
    combination of these distributions.

    Returns
    -------
    stationary_dists : array_like(float, ndim=2)
        Array containing the stationary distributions as its rows.

    """
    return MarkovChain(P).stationary_distributions


def mc_sample_path(P, init=0, sample_size=1000, random_state=None):
    """
    Generates one sample path from the Markov chain represented by
    (n x n) transition matrix P on state space S = {{0,...,n-1}}.

    Parameters
    ----------
    P : array_like(float, ndim=2)
        A Markov transition matrix.

    init : array_like(float ndim=1) or scalar(int), optional(default=0)
        If init is an array_like, then it is treated as the initial
        distribution across states.  If init is a scalar, then it
        treated as the deterministic initial state.

    sample_size : scalar(int), optional(default=1000)
        The length of the sample path.

    random_state : int or np.random.RandomState/Generator, optional
        Random seed (integer) or np.random.RandomState or Generator
        instance to set the initial state of the random number generator
        for reproducibility. If None, a randomly initialized RandomState
        is used.

    Returns
    -------
    X : array_like(int, ndim=1)
        The simulation of states.

    """
    random_state = check_random_state(random_state)

    if isinstance(init, numbers.Integral):
        X_0 = init
    else:
        cdf0 = np.cumsum(init)
        u_0 = random_state.random()
        X_0 = searchsorted(cdf0, u_0)

    mc = MarkovChain(P)
    return mc.simulate(ts_length=sample_size, init=X_0,
                       random_state=random_state)



"""
Methods for computing pure Nash equilibria of a normal form game.
(For now, only brute force method is supported)

"""

import numpy as np


def pure_nash_brute(g, tol=None):
    """
    Find all pure Nash equilibria of a normal form game by brute force.

    Parameters
    ----------
    g : NormalFormGame
    tol : scalar(float), optional(default=None)
        Tolerance level used in determining best responses. If None,
        default to the value of the `tol` attribute of `g`.

    Returns
    -------
    NEs : list(tuple(int))
        List of tuples of Nash equilibrium pure actions.
        If no pure Nash equilibrium is found, return empty list.

    Examples
    --------
    Consider the "Prisoners' Dilemma" game:

    >>> PD_bimatrix = [[(1, 1), (-2, 3)],
    ...                [(3, -2), (0, 0)]]
    >>> g_PD = NormalFormGame(PD_bimatrix)
    >>> pure_nash_brute(g_PD)
    [(1, 1)]

    If we consider the "Matching Pennies" game, which has no pure nash
    equilibrium:

    >>> MP_bimatrix = [[(1, -1), (-1, 1)],
    ...                [(-1, 1), (1, -1)]]
    >>> g_MP = NormalFormGame(MP_bimatrix)
    >>> pure_nash_brute(g_MP)
    []

    """
    return list(pure_nash_brute_gen(g, tol=tol))


def pure_nash_brute_gen(g, tol=None):
    """
    Generator version of `pure_nash_brute`.

    Parameters
    ----------
    g : NormalFormGame
    tol : scalar(float), optional(default=None)
        Tolerance level used in determining best responses. If None,
        default to the value of the `tol` attribute of `g`.

    Yields
    ------
    out : tuple(int)
        Tuple of Nash equilibrium pure actions.

    """
    for a in np.ndindex(*g.nums_actions):
        if g.is_nash(a, tol=tol):
            yield a



"""
Tests for discrete_rv.py

"""

import numpy as np
from numpy.testing import assert_allclose, assert_
import pytest
from quantecon import DiscreteRV


class TestDiscreteRV:

    @classmethod
    def setup_method(cls):
        x = np.random.rand(10)
        x /= x.sum()
        # make sure it sums to 1
        cls.x = x
        cls.drv = DiscreteRV(cls.x)

    def test_Q_updates(self):
        "discrete_rv: Q attributes updates on q change?"
        Q_init = np.copy(self.drv.Q)

        # change q, see if Q updates
        x = np.random.rand(10)
        x /= x.sum()
        self.drv.q = x
        Q_after = self.drv.Q

        # should be different
        assert_(not np.allclose(Q_init, Q_after))

        # clean up: reset values
        self.drv.q = self.x

        # now we should have our original Q back
        assert_allclose(Q_init, self.drv.Q)

    def test_Q_end_1(self):
        "discrete_rv: Q sums to 1"
        assert_(self.drv.Q[-1] - 1.0 < 1e-10)

    @pytest.mark.slow
    def test_draw_lln(self):
        "discrete_rv: lln satisfied?"
        draws = self.drv.draw(1000000)
        bins = np.arange(self.drv.q.size+1)
        freqs, _ = np.histogram(draws, bins=bins, density=True)
        assert_allclose(freqs, self.drv.q, atol=1e-2)

    def test_draw_with_seed(self):
        x = np.array([0.03326189, 0.60713005, 0.84514831, 0.28493183,
                      0.12393182, 0.35308009, 0.70371579, 0.81728178,
                      0.21294538, 0.05358209])
        draws = DiscreteRV(x).draw(k=10, random_state=5)

        expected_output = np.array([1, 2, 1, 2, 1, 1, 2, 1, 1, 1])

        assert_allclose(draws, expected_output)



"""
Tests for pure_nash.py

"""

import numpy as np
import itertools
from numpy.testing import assert_
from quantecon.game_theory import NormalFormGame, pure_nash_brute


class TestPureNashBruteForce():
    def setup_method(self):
        self.game_dicts = []

        # Matching Pennies game with no pure nash equilibrium
        MP_bimatrix = [[(1, -1), (-1, 1)],
                       [(-1, 1), (1, -1)]]
        MP_NE = []

        # Prisoners' Dilemma game with one pure nash equilibrium
        PD_bimatrix = [[(1, 1), (-2, 3)],
                       [(3, -2), (0, 0)]]
        PD_NE = [(1, 1)]

        # Battle of the Sexes game with two pure nash equilibria
        BoS_bimatrix = [[(3, 2), (1, 0)],
                        [(0, 1), (2, 3)]]
        BoS_NE = [(0, 0), (1, 1)]

        # Unanimity Game with more than two players
        N = 4
        a, b = 1, 2
        g_Unanimity = NormalFormGame((2,)*N)
        g_Unanimity[(0,)*N] = (a,)*N
        g_Unanimity[(1,)*N] = (b,)*N

        Unanimity_NE = [(0,)*N]
        for k in range(2, N-2+1):
            for ind in itertools.combinations(range(N), k):
                a = np.ones(N, dtype=int)
                a[list(ind)] = 0
                Unanimity_NE.append(tuple(a))
        Unanimity_NE.append((1,)*N)

        for bimatrix, NE in zip([MP_bimatrix, PD_bimatrix, BoS_bimatrix],
                                [MP_NE, PD_NE, BoS_NE]):
            d = {'g': NormalFormGame(bimatrix),
                 'NEs': NE}
            self.game_dicts.append(d)
        self.game_dicts.append({'g': g_Unanimity,
                                'NEs': Unanimity_NE})

    def test_brute_force(self):
        for d in self.game_dicts:
            assert_(sorted(pure_nash_brute(d['g'])) == sorted(d['NEs']))

    def test_tol(self):
        # Prisoners' Dilemma game with one NE and one epsilon NE
        epsilon = 1e-08

        PD_bimatrix = [[(1, 1), (-2, 1 + epsilon)],
                       [(1 + epsilon, -2), (0, 0)]]

        NEs = [(1, 1)]
        epsilon_NEs = [(1, 1), (0, 0)]

        g = NormalFormGame(PD_bimatrix)
        for tol, answer in zip([0, epsilon], [NEs, epsilon_NEs]):
            assert_(sorted(pure_nash_brute(g, tol=tol)) == sorted(answer))



"""
Utilities to Support Random State Infrastructure
"""

import numpy as np
import numbers

# To be called as util.rng_integers
from scipy._lib._util import rng_integers  # noqa: F401


# Random States

def check_random_state(seed):
    """
    Turn `seed` into a `np.random.RandomState` instance.

    Parameters
    ----------
    seed : None, int, `np.random.RandomState`, or `np.random.Generator`
        If None, the `np.random.RandomState` singleton is returned. If
        `seed` is an int, a new ``RandomState`` instance seeded with
        `seed` is returned. If `seed` is already a `RandomState` or
        `Generator` instance, then that instance is returned.

    Returns
    -------
    `np.random.RandomState` or `np.random.Generator`
        Random number generator.

    Notes
    -----
    This code was originally sourced from scikit-learn.

    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, (np.random.RandomState, np.random.Generator)):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)
